// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file ComponentTraverseInfo.h */

#ifndef AMDIS_COMPONENTTRAVERSEINFO_H
#define AMDIS_COMPONENTTRAVERSEINFO_H

#include <vector>
#include "Global.h"
#include "FiniteElemSpace.h"

namespace AMDiS {

  class SingleComponentInfo
  {      
  public:
    SingleComponentInfo()
      : rowFeSpace(NULL),
	colFeSpace(NULL),
	auxFeSpaces(0),
	status(0)
      {}
    
    void setFeSpace(FiniteElemSpace *row, FiniteElemSpace *col = NULL) 
    {
      rowFeSpace = row;
      colFeSpace = col;
    }
    
    void setAuxFeSpaces(std::vector<const FiniteElemSpace*> feSpaces) 
    {
      auxFeSpaces = feSpaces;
    }

    void addAuxFeSpace(const FiniteElemSpace *fe) 
    {
      auxFeSpaces.push_back(fe);
    }
    
    bool hasFeSpace() 
    {
      return rowFeSpace != NULL;
    }

    void updateStatus();
    
    int getNumAuxFeSpaces() 
    {
      return auxFeSpaces.size();
    }
    
    FiniteElemSpace *getRowFeSpace() 
    {
      return rowFeSpace;
    }
    
    FiniteElemSpace *getColFeSpace() 
    {
      return colFeSpace;
    }
    
    const FiniteElemSpace *getAuxFeSpace(int i) 
    {
      return ((i < static_cast<int>(auxFeSpaces.size())) ? auxFeSpaces[i] : NULL);
    }

    void setAuxFeSpace(const FiniteElemSpace* fe, int pos) 
    {
      auxFeSpaces[pos] = fe;
    }

    int getStatus() 
    {
      return status;
    }
    
  protected:      
    FiniteElemSpace *rowFeSpace;
    
    FiniteElemSpace *colFeSpace;
    
    std::vector<const FiniteElemSpace*> auxFeSpaces;

    /// Status of the component, see the possible status flags below.
    int status;

  public:
    /// Single component status flag: empty component, no fe spaces
    static const int EMPTY;

    /// Single component status flag: row = col, no aux
    static const int EQ_SPACES_NO_AUX;

    /// Single component status flag: row = col = aux
    static const int EQ_SPACES_WITH_AUX;

    /// Single component status flag: row = col, different aux
    static const int EQ_SPACES_WITH_DIF_AUX;

    /// Single component status flag: row, col, no aux
    static const int DIF_SPACES_NO_AUX;

    /// Single component status flag: row, col, aux either equal to row or to col
    static const int DIF_SPACES_WITH_AUX;

    /// Single component status flag: row, col, aux (at least 3 different fe spaces)
    static const int DIF_SPACES_WITH_DIF_AUX;
  };

  
  class ComponentTraverseInfo
  {
  public: 
    ComponentTraverseInfo(int n) 
      : nComponents(n)
    {
      resize(n);
    }

    void resize(int n) 
    {
      nComponents = n;

      matrixComponents.resize(n);
      vectorComponents.resize(n);

      for (int i = 0; i < n; i++)
	matrixComponents[i].resize(n);
    }

    void updateStatus() 
    {
      for (int i = 0; i < nComponents; i++) {
	for (int j = 0; j < nComponents; j++)
	  matrixComponents[i][j].updateStatus();

	vectorComponents[i].updateStatus();
      }
    }

    SingleComponentInfo &getMatrix(int row, int col) 
    {
      return matrixComponents[row][col];
    }

    SingleComponentInfo &getVector(int row) 
    {
      return vectorComponents[row];
    }

    int getStatus(int row, int col) 
    {
      return matrixComponents[row][col].getStatus();
    }

    int getStatus(int row) 
    {
      return vectorComponents[row].getStatus();
    }

    const FiniteElemSpace* getAuxFeSpace(int row, int col) 
    {
      return matrixComponents[row][col].getAuxFeSpace(0);
    }

    const FiniteElemSpace* getAuxFeSpace(int row) 
    {
      return vectorComponents[row].getAuxFeSpace(0);
    }
    
    /** \brief
     * Returns the row FE space for a given row number, i.e., the FE space
     * of the diagonal matrix.
     *
     * \param[in]  row   Row number of the matrix line for which the FE space
     *                   should be returned.
     */    
    const FiniteElemSpace* getRowFeSpace(int row);


    /** \brief
     * Returns the non row FE space for a given row number. This is either the
     * col FE space of an off diagonal matrix or the aux fe space of another
     * matrix in the row or of the right hand side vector. If there is no non row
     * FE space, this function returns a null pointer.
     *
     * \param[in]  row   Row number of the matrix line for which the non FE space
     *                   should be returned.
     */
    const FiniteElemSpace* getNonRowFeSpace(int row);

  protected:
    int nComponents;

    std::vector<std::vector<SingleComponentInfo> > matrixComponents;

    std::vector<SingleComponentInfo> vectorComponents;
  };

}

#endif
